/* The sample code computes the L2-projection of a given function into the space of DG
basis functions described in DG_elements. The mesh data is provided in the directory 
../Meshes. The function Mesh_Gen will read this data and produce the mesh. NOTE: might
have to change the read path specified in Mesh_Gen.*/

static char help[] = "Minimal working example of error in global matrix assembly.\n\n";

#include <petscsys.h>
#include <petscvec.h>
#include <petscmat.h>
#include <petscksp.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

void Mesh_Gen(int ref,
			  int nNodes,
			  int nElems,
			  PetscScalar* Coordinates, 
			  PetscInt* Elements);

void get_Mesh_Parameters(int ref,
	                     int* nNodes,
	                     int* nElems);

void DG_Solve(int pdeg,
			  int ref);


void L2_Projection(int pdeg, 
				   int NLoc,
				   int nNodes,
				   int nElems,
				   PetscScalar* Coordinates, 
				   PetscInt*Elements,
				   Vec usol0);

void DG_local_mass(int pdeg,
				   int NLoc,
				   int K,
				   PetscScalar* Mlocal,
				   PetscScalar* rhs_local,
				   PetscScalar* Coordinates,
				   PetscInt* Elements);

void get_Element(int K,
				 double *area,
				 int* Elements,
				 double* BK,
				 double* bk,
				 double* Coordinates);

void DG_element(int pdeg,
				int NLoc,
				double x,
				double y,
				double* BK,
				double* value_basis,
				double* grad_basis);

void init_zeros(int m,
				int n,
				double* A);

void map_Quadrature(double x,
                    double y,
                    double *xglob1,
                    double *xglob2,
                    double* BK,
                    double* bk);

void QuadRule_2D(int ng,
		         double *wg,
	             double *xg);

void exact_Solution(double x,
					double y,
					double* u_exact);

void get_2x2_inv(double* inv_B,
			     double* B);


/* This is the function to perform the L2-projection on (given by *u_exact) */
void exact_Solution(double x,
					double y,
					double* u_exact)
{
	*u_exact = x*(x-1)*y*(y-1);
}


/* main driver */
int main(int argc, 
		 char **argv)
{

	PetscErrorCode ierr;

	int ref;
	int pdeg; 

	ierr = PetscInitialize(&argc,&argv,(char*)0,help); CHKERRQ(ierr);

	/* number of refinements on the mesh */
	ref = 5;

	/* degree of basis functions */
	pdeg = 1;

	DG_Solve(pdeg, ref);

	ierr = PetscFinalize();

	return 0;

}


/* Get the number of nodes and element for each mesh. */
void get_Mesh_Parameters(int ref,
	                     int* nNodes,
	                     int* nElems)
{

	*nNodes = 0;
	*nElems = 0;

	/* mesh parameters */
	if (ref == 0)
	{
		*nNodes = 4;
		*nElems = 2;
	}
	else if (ref == 1)
	{
		*nNodes = 9;
		*nElems = 8;
	}
	else if (ref == 2)
	{
		*nNodes = 25;
		*nElems = 32;
	}
	else if (ref == 3)
	{
		*nNodes = 81;
		*nElems = 128;
	}
	else if (ref == 4)
	{
		*nNodes = 289;
		*nElems = 512;
	}
	else if (ref == 5)
	{
		*nNodes = 1089;
		*nElems = 2048;
	}
	else if (ref == 6)
	{
		*nNodes = 4225;
		*nElems = 8192;
	}

}



void DG_Solve(int pdeg,
			  int ref)
{

	int nNodes, nElems, NLoc;

	PetscScalar* Coordinates;
	PetscInt* Elements;

	Vec usol0;

	NLoc = (pdeg+1)*(pdeg+2)/2;

	/* mesh parameters */
	get_Mesh_Parameters(ref, &nNodes, &nElems);

	PetscMalloc(sizeof(PetscScalar)*nNodes*2, &Coordinates);
	PetscMalloc(sizeof(PetscInt)*nElems*3, &Elements);
	
	/* usol0 will be the vector of coefficients for the L2-projection of the function into 
	the DG space */
	VecCreateSeq(PETSC_COMM_SELF, NLoc*nElems, &usol0);

	/* generate the mesh */
	Mesh_Gen(ref, nNodes, nElems, Coordinates, Elements);

	/* get initial guess via L2-projection */
	L2_Projection(pdeg, NLoc, nNodes, nElems, Coordinates, Elements, usol0);

	//VecView(usol0, PETSC_VIEWER_STDOUT_WORLD);

	PetscFree(Coordinates);
	PetscFree(Elements);

	VecDestroy(&usol0);

}



/* Read mesh data from .dat files and save to the prescribed arrays. Current implementation
only supports meshes on the unit square.

	Output:
		Coordinates = nNodesx2 array of coordinates of mesh nodes
		Elements	= nElemsx3 array of the nodes that make up each element
*/

#define Coordinates(i,j) Coordinates[ (j-1) + 2*(i-1) ]
#define Elements(i,j)    Elements[ (j-1) + 3*(i-1) ]

void Mesh_Gen(int ref,
			  int nNodes,
			  int nElems,
			  PetscScalar* Coordinates, 
			  PetscInt* Elements)
{

	double x;
	PetscInt y;

	char str[80];
	char c[80];
	char e2[80];

	int i, j;

	FILE* coord  = NULL;
	FILE* elem   = NULL;

	/* specifies read path; may need to change on your machine */
	sprintf(str, "../Meshes/Mesh%i", ref);

	strcpy(c, str);  strcat(c, "Coord.dat");
	strcpy(e2, str); strcat(e2, "Elem.dat");

	coord  = fopen(c, "r");
	elem   = fopen(e2, "r");

	for (i = 1; i <= nNodes; ++i)
	{
		for (j = 1; j <= 2; ++j)
		{
			fscanf(coord, "%lf", &x);
			Coordinates(i,j) = x;
		}
	}

	for (i = 1; i <= nElems; ++i)
	{
		for (j = 1; j <= 3; ++j)
		{
			fscanf(elem, "%d", &y);
			Elements(i,j) = y;
		}
	}

	fclose(coord);
	fclose(elem);

}


/* compute the L2-projection */
void L2_Projection(int pdeg, 
				   int NLoc,
				   int nNodes,
				   int nElems,
				   PetscScalar* Coordinates, 
				   PetscInt*Elements,
				   Vec usol0)
{

	int i, k;

	PetscInt* ie;
	PetscInt* je;
	
	PetscScalar* Mlocal;
	PetscScalar* rhs_local;

	Mat Mglobal;

	KSP ksp;
	PC pc;

	MatCreateSeqAIJ(PETSC_COMM_SELF, NLoc*nElems, NLoc*nElems, NLoc, 
				    PETSC_NULL, &Mglobal);

	PetscMalloc(sizeof(PetscScalar)*NLoc*NLoc, &Mlocal);
	PetscMalloc(sizeof(PetscScalar)*NLoc, &rhs_local);
	PetscMalloc(sizeof(PetscInt)*NLoc, &ie);
	PetscMalloc(sizeof(PetscInt)*NLoc, &je);

	/* assemble volume contributions to mass matrix */
	for (k = 1; k <= nElems; ++k)
	{

		for (i = 0; i < NLoc; ++i)
		{
			ie[i] = i + NLoc*(k-1);
			je[i] = i + NLoc*(k-1);
		}

		DG_local_mass(pdeg, NLoc, k, Mlocal, rhs_local, Coordinates, Elements);

		MatSetValues(Mglobal, NLoc, ie, NLoc, je, Mlocal, ADD_VALUES);
		VecSetValues(usol0, NLoc, ie, rhs_local, ADD_VALUES);
	}

	MatAssemblyBegin(Mglobal, MAT_FINAL_ASSEMBLY);
	MatAssemblyEnd(Mglobal, MAT_FINAL_ASSEMBLY);

	MatView(Mglobal, PETSC_VIEWER_STDOUT_WORLD);

	/* solve the linear system */
	KSPCreate(PETSC_COMM_WORLD, &ksp);
	KSPSetOperators(ksp, Mglobal, Mglobal, DIFFERENT_NONZERO_PATTERN);
	KSPGetPC(ksp, &pc);

	PCSetType(pc, PCLU);
	KSPSetTolerances(ksp, 1.e-5, PETSC_DEFAULT, PETSC_DEFAULT, PETSC_DEFAULT);

	KSPSolve(ksp, usol0, usol0);

	MatDestroy(&Mglobal);
	KSPDestroy(&ksp);

	PetscFree(Mlocal);
	PetscFree(rhs_local);
	PetscFree(ie);
	PetscFree(je);

}


/* compute the local mass matrix on the element with global number K */
#define Mlocal(i,j) Mlocal[ (j-1) + NLoc*(i-1) ]
#define xg(i,j)     xg[ (j-1) + 2*(i-1) ]

void DG_local_mass(int pdeg,
				   int NLoc,
				   int K,
				   PetscScalar* Mlocal,
				   PetscScalar* rhs_local,
				   PetscScalar* Coordinates,
				   PetscInt* Elements)
{

	int i, j, k, ng;

	double area;
	double L2Proj;
	double x, y;

	double* wg;
	double* xg;
	double* BK;
	double* bk;
	double* value_basis;
	double* grad_basis;

	/* number of quadrature points; can set to 7 or 25 */
	ng = 25;

	wg = (double*) malloc(sizeof(double)*ng);
	xg = (double*) malloc(sizeof(double)*ng*2);

	BK = (double *) malloc(sizeof(double)*2*2);
	bk = (double *) malloc(sizeof(double)*2);

	value_basis = (double*) malloc(sizeof(double)*NLoc);
	grad_basis  = (double*) malloc(sizeof(double)*NLoc*2);

	QuadRule_2D(ng, wg, xg);

	get_Element(K, &area, Elements, BK, bk, Coordinates);

	init_zeros(NLoc, NLoc, Mlocal);
	init_zeros(NLoc, 1, rhs_local);

	for (k = 1; k <= ng; ++k)
	{
		DG_element(pdeg, NLoc, xg(k,1), xg(k,2), BK, value_basis, grad_basis);
		map_Quadrature(xg(k,1), xg(k,2), &x, &y, BK, bk);

		for (i = 1; i <= NLoc; ++i)
		{
			for (j = 1; j <= NLoc; ++j)
			{
				Mlocal(i,j) += area*wg[k-1]*value_basis[j-1]*value_basis[i-1];
			}

			exact_Solution(x, y, &L2Proj);
			rhs_local[i-1] += area*wg[k-1]*L2Proj*value_basis[i-1];
		}
	}

	free(wg);
	free(xg);
	free(BK);
	free(bk);
	free(value_basis);
	free(grad_basis);

}


/* get affine transformation on the element K */
#define BK(i,j)			 BK[ (j-1) + 2*(i-1) ]
#define Elements(i,j)    Elements[ (j-1) + 3*(i-1) ]
#define Coordinates(i,j) Coordinates[ (j-1) + 2*(i-1) ]

void get_Element(int K,
				 double *area,
				 int* Elements,
				 double* BK,
				 double* bk,
				 double* Coordinates)
{

	int n1;
	int n2;
	int n3;

	n1 = Elements(K,1);
	n2 = Elements(K,2);
	n3 = Elements(K,3);

 	bk[0] = Coordinates(n1,1);
	bk[1] = Coordinates(n1,2);

	BK(1,1) = Coordinates(n2,1) - Coordinates(n1,1);
	BK(2,1) = Coordinates(n2,2) - Coordinates(n1,2);
	BK(1,2) = Coordinates(n3,1) - Coordinates(n1,1);
	BK(2,2) = Coordinates(n3,2) - Coordinates(n1,2);

	/* determinant of the Jacobian, twice the area of the element */
	*area = BK(1,1)*BK(2,2) - BK(1,2)*BK(2,1);

}


/* 2D quadrature rule on the triangle with coordinates (0,0), (1,0), (0,1) */
#define xg(i,j) xg[ (j-1) + 2*(i-1) ]

void QuadRule_2D(int ng,
		         double *wg,
	             double *xg)
{

	if (ng == 7)
	{
		wg[0] = 0.112500000000;
        wg[1] = 0.062969590272;
        wg[2] = 0.062969590272;
        wg[3] = 0.062969590272;
        wg[4] = 0.066197076394;
        wg[5] = 0.066197076394;
        wg[6] = 0.066197076394;
        
        xg(1,1) = 0.333333333333;
        xg(1,2) = 0.333333333333;
        xg(2,1) = 0.101286507323;
        xg(2,2) = 0.101286507323;
        xg(3,1) = 0.797426985353;
        xg(3,2) = 0.101286507323;
        xg(4,1) = 0.101286507323;
        xg(4,2) = 0.797426985353;
        xg(5,1) = 0.470142064105;
        xg(5,2) = 0.470142064105;
        xg(6,1) = 0.059715871789;
        xg(6,2) = 0.470142064105;
        xg(7,1) = 0.470142064105;
        xg(7,2) = 0.059715871789;
	}
    else if (ng == 25)
    {
        wg[0] = 0.090817990382754/2.0;
        wg[1] = 0.036725957756467/2.0;
        wg[2] = 0.036725957756467/2.0;
        wg[3] = 0.036725957756467/2.0;
        wg[4] = 0.045321059435528/2.0;
        wg[5] = 0.045321059435528/2.0;
        wg[6] = 0.045321059435528/2.0;
        
        wg[7] = 0.072757916845420/2.0;
        wg[8] = 0.072757916845420/2.0;
        wg[9] = 0.072757916845420/2.0;
        wg[10] = 0.072757916845420/2.0;
        wg[11] = 0.072757916845420/2.0;
        wg[12] = 0.072757916845420/2.0;
        
        wg[13] = 0.028327242531057/2.0;
        wg[14] = 0.028327242531057/2.0;
        wg[15] = 0.028327242531057/2.0;
        wg[16] = 0.028327242531057/2.0;
        wg[17] = 0.028327242531057/2.0;
        wg[18] = 0.028327242531057/2.0;
        
        wg[19] = 0.009421666963733/2.0;
        wg[20] = 0.009421666963733/2.0;
        wg[21] = 0.009421666963733/2.0;
        wg[22] = 0.009421666963733/2.0;
        wg[23] = 0.009421666963733/2.0;
        wg[24] = 0.009421666963733/2.0;

        xg(1,1) = 0.333333333333333;
        xg(1,2) = 0.333333333333333;
        xg(2,1) = 0.485577633383657;
        xg(2,2) = 0.485577633383657;
        xg(3,1) = 0.028844733232685;
        xg(3,2) = 0.485577633383657;
        xg(4,1) = 0.485577633383657;
        xg(4,2) = 0.028844733232685;
        xg(5,1) = 0.109481575485037;
        xg(5,2) = 0.109481575485037;
        xg(6,1) = 0.781036849029926;
        xg(6,2) = 0.109481575485037;
        xg(7,1) = 0.109481575485037;
        xg(7,2) = 0.781036849029926;
            
        xg(8,1) = 0.307939838764121;
        xg(8,2) = 0.550352941820999;
        xg(9,1) = 0.550352941820999;
        xg(9,2) = 0.307939838764121;
        xg(10,1) = 0.141707219414880;
        xg(10,2) = 0.550352941820999;
        xg(11,1) = 0.550352941820999;
        xg(11,2) = 0.141707219414880;
        xg(12,1) = 0.307939838764121;
        xg(12,2) = 0.141707219414880;
        xg(13,1) = 0.141707219414880;
        xg(13,2) = 0.307939838764121;
            
        xg(14,1) = 0.246672560639903;
        xg(14,2) = 0.728323904597411;
        xg(15,1) = 0.728323904597411;
        xg(15,2) = 0.246672560639903;
        xg(16,1) = 0.025003534762686;
        xg(16,2) = 0.728323904597411;
        xg(17,1) = 0.728323904597411;
        xg(17,2) = 0.025003534762686;
        xg(18,1) = 0.246672560639903;
        xg(18,2) = 0.025003534762686;
        xg(19,1) = 0.025003534762686;
        xg(19,2) = 0.246672560639903;
            
        xg(20,1) = 0.066803251012200;
        xg(20,2) = 0.923655933587500;
        xg(21,1) = 0.923655933587500;
        xg(21,2) = 0.066803251012200;
        xg(22,1) = 0.066803251012200;
        xg(22,2) = 0.009540815400299;
        xg(23,1) = 0.009540815400299;
        xg(23,2) = 0.066803251012200;
        xg(24,1) = 0.923655933587500; 
        xg(24,2) = 0.009540815400299;
        xg(25,1) = 0.009540815400299;
        xg(25,2) = 0.923655933587500;
    }

}


/* map a given quadrature point from the reference triangle to the global element
using the affine transformationd described by BK and bk */
#define BK(i,j) BK[ (j-1) + 2*(i-1) ]

void map_Quadrature(double x,
                    double y,
                    double *xglob1,
                    double *xglob2,
                    double* BK,
                    double* bk)
{

    *xglob1 = BK(1,1)*x + BK(1,2)*y + bk[0];
    *xglob2 = BK(2,1)*x + BK(2,2)*y + bk[1];

}



/* initialize a given matrix to zeros */
#define A(i,j) A[ (j-1) + n*(i-1) ]

void init_zeros(int m,
				int n,
				double* A)
{

	int i, j;

	for (i = 1; i <= m; ++i)
	{
		for (j = 1; j <= n; ++j)
		{
			A(i,j) = 0.0;
		}
	}

}


/* basis functions */
#define grad_basis(i,j) grad_basis[ (j-1) + 2*(i-1) ]
#define BK(i,j)			BK[ (j-1) + 2*(i-1) ]
#define inv_BK(i,j)		inv_BK[ (j-1) + 2*(i-1)]
#define grad_basis_temp(i,j) grad_basis_temp[ (j-1) + 2*(i-1) ]

void DG_element(int pdeg,
				int NLoc,
				double x,
				double y,
				double* BK,
				double* value_basis,
				double* grad_basis)
{

	double a;
	double b;
	double c;
	double d;
	double det;

	double* inv_BK;
	double* grad_basis_temp;

	inv_BK = (double*) malloc(sizeof(double)*2*2);

	grad_basis_temp = (double*) malloc(sizeof(double)*NLoc*2);

	a = BK(1,1);
	b = BK(1,2);
	c = BK(2,1);
	d = BK(2,2);

	/* determine of the transpose of BK */
	det = a*d - b*c;

	inv_BK(1,1) = 1/det*d;
	inv_BK(1,2) = 1/det*-c;
	inv_BK(2,2) = 1/det*a;
	inv_BK(2,1) = 1/det*-b;

	if (pdeg == 1)
	{
		value_basis[0] = 1.0;
		value_basis[1] = x;
		value_basis[2] = y;

		grad_basis_temp(1,1) = 0.0;
		grad_basis_temp(1,2) = 0.0;
		grad_basis_temp(2,1) = 1.0;
		grad_basis_temp(2,2) = 0.0;
		grad_basis_temp(3,1) = 0.0;
		grad_basis_temp(3,2) = 1.0;

		grad_basis(1,1) = inv_BK(1,1)*grad_basis_temp(1,1) + inv_BK(1,2)*grad_basis_temp(1,2);
		grad_basis(1,2) = inv_BK(2,1)*grad_basis_temp(1,1) + inv_BK(2,2)*grad_basis_temp(1,2);
		grad_basis(2,1) = inv_BK(1,1)*grad_basis_temp(2,1) + inv_BK(1,2)*grad_basis_temp(2,2);
		grad_basis(2,2) = inv_BK(2,1)*grad_basis_temp(2,1) + inv_BK(2,2)*grad_basis_temp(2,2);
		grad_basis(3,1) = inv_BK(1,1)*grad_basis_temp(3,1) + inv_BK(1,2)*grad_basis_temp(3,2);
		grad_basis(3,2) = inv_BK(2,1)*grad_basis_temp(3,1) + inv_BK(2,2)*grad_basis_temp(3,2);
	}
	else if (pdeg == 2)
	{
		value_basis[0] = 1.0;
		value_basis[1] = x;
		value_basis[2] = y;
		value_basis[3] = x*x;
		value_basis[4] = x*y;
		value_basis[5] = y*y;

		grad_basis_temp(1,1) = 0.0;
		grad_basis_temp(1,2) = 0.0;
		grad_basis_temp(2,1) = 1.0;
		grad_basis_temp(2,2) = 0.0;
		grad_basis_temp(3,1) = 0.0;
		grad_basis_temp(3,2) = 1.0;

		grad_basis_temp(4,1) = 2*x;
		grad_basis_temp(4,2) = 0.0;
		grad_basis_temp(5,1) = y;
		grad_basis_temp(5,2) = x;
		grad_basis_temp(6,1) = 0.0;
		grad_basis_temp(6,2) = 2*y;

		grad_basis(1,1) = inv_BK(1,1)*grad_basis_temp(1,1) + inv_BK(1,2)*grad_basis_temp(1,2);
		grad_basis(1,2) = inv_BK(2,1)*grad_basis_temp(1,1) + inv_BK(2,2)*grad_basis_temp(1,2);
		grad_basis(2,1) = inv_BK(1,1)*grad_basis_temp(2,1) + inv_BK(1,2)*grad_basis_temp(2,2);
		grad_basis(2,2) = inv_BK(2,1)*grad_basis_temp(2,1) + inv_BK(2,2)*grad_basis_temp(2,2);
		grad_basis(3,1) = inv_BK(1,1)*grad_basis_temp(3,1) + inv_BK(1,2)*grad_basis_temp(3,2);
		grad_basis(3,2) = inv_BK(2,1)*grad_basis_temp(3,1) + inv_BK(2,2)*grad_basis_temp(3,2);

		grad_basis(4,1) = inv_BK(1,1)*grad_basis_temp(4,1) + inv_BK(1,2)*grad_basis_temp(4,2);
		grad_basis(4,2) = inv_BK(2,1)*grad_basis_temp(4,1) + inv_BK(2,2)*grad_basis_temp(4,2);
		grad_basis(5,1) = inv_BK(1,1)*grad_basis_temp(5,1) + inv_BK(1,2)*grad_basis_temp(5,2);
		grad_basis(5,2) = inv_BK(2,1)*grad_basis_temp(5,1) + inv_BK(2,2)*grad_basis_temp(5,2);
		grad_basis(6,1) = inv_BK(1,1)*grad_basis_temp(6,1) + inv_BK(1,2)*grad_basis_temp(6,2);
		grad_basis(6,2) = inv_BK(2,1)*grad_basis_temp(6,1) + inv_BK(2,2)*grad_basis_temp(6,2);
	}
	else if (pdeg == 3)
	{
		value_basis[0] = 1.0;
		value_basis[1] = x;
		value_basis[2] = y;
		value_basis[3] = x*x;
		value_basis[4] = x*y;
		value_basis[5] = y*y;
		value_basis[6] = x*x*x;
		value_basis[7] = x*x*y;
		value_basis[8] = x*y*y;
		value_basis[9] = y*y*y;

		grad_basis_temp(1,1) = 0.0;
		grad_basis_temp(1,2) = 0.0;
		grad_basis_temp(2,1) = 1.0;
		grad_basis_temp(2,2) = 0.0;
		grad_basis_temp(3,1) = 0.0;
		grad_basis_temp(3,2) = 1.0;

		grad_basis_temp(4,1) = 2*x;
		grad_basis_temp(4,2) = 0.0;
		grad_basis_temp(5,1) = y;
		grad_basis_temp(5,2) = x;
		grad_basis_temp(6,1) = 0.0;
		grad_basis_temp(6,2) = 2*y;

		grad_basis_temp(7,1) = 3*x*x;
		grad_basis_temp(7,2) = 0.0;
		grad_basis_temp(8,1) = 2*x*y;
		grad_basis_temp(8,2) = x*x;
		grad_basis_temp(9,1) = y*y;
		grad_basis_temp(9,2) = 2*x*y;
		grad_basis_temp(10,1) = 0.0;
		grad_basis_temp(10,2) = 3*y*y;

		grad_basis(1,1) = inv_BK(1,1)*grad_basis_temp(1,1) + inv_BK(1,2)*grad_basis_temp(1,2);
		grad_basis(1,2) = inv_BK(2,1)*grad_basis_temp(1,1) + inv_BK(2,2)*grad_basis_temp(1,2);
		grad_basis(2,1) = inv_BK(1,1)*grad_basis_temp(2,1) + inv_BK(1,2)*grad_basis_temp(2,2);
		grad_basis(2,2) = inv_BK(2,1)*grad_basis_temp(2,1) + inv_BK(2,2)*grad_basis_temp(2,2);
		grad_basis(3,1) = inv_BK(1,1)*grad_basis_temp(3,1) + inv_BK(1,2)*grad_basis_temp(3,2);
		grad_basis(3,2) = inv_BK(2,1)*grad_basis_temp(3,1) + inv_BK(2,2)*grad_basis_temp(3,2);

		grad_basis(4,1) = inv_BK(1,1)*grad_basis_temp(4,1) + inv_BK(1,2)*grad_basis_temp(4,2);
		grad_basis(4,2) = inv_BK(2,1)*grad_basis_temp(4,1) + inv_BK(2,2)*grad_basis_temp(4,2);
		grad_basis(5,1) = inv_BK(1,1)*grad_basis_temp(5,1) + inv_BK(1,2)*grad_basis_temp(5,2);
		grad_basis(5,2) = inv_BK(2,1)*grad_basis_temp(5,1) + inv_BK(2,2)*grad_basis_temp(5,2);
		grad_basis(6,1) = inv_BK(1,1)*grad_basis_temp(6,1) + inv_BK(1,2)*grad_basis_temp(6,2);
		grad_basis(6,2) = inv_BK(2,1)*grad_basis_temp(6,1) + inv_BK(2,2)*grad_basis_temp(6,2);

		grad_basis(7,1) = inv_BK(1,1)*grad_basis_temp(7,1) + inv_BK(1,2)*grad_basis_temp(7,2);
		grad_basis(7,2) = inv_BK(2,1)*grad_basis_temp(7,1) + inv_BK(2,2)*grad_basis_temp(7,2);
		grad_basis(8,1) = inv_BK(1,1)*grad_basis_temp(8,1) + inv_BK(1,2)*grad_basis_temp(8,2);
		grad_basis(8,2) = inv_BK(2,1)*grad_basis_temp(8,1) + inv_BK(2,2)*grad_basis_temp(8,2);
		grad_basis(9,1) = inv_BK(1,1)*grad_basis_temp(9,1) + inv_BK(1,2)*grad_basis_temp(9,2);
		grad_basis(9,2) = inv_BK(2,1)*grad_basis_temp(9,1) + inv_BK(2,2)*grad_basis_temp(9,2);
		grad_basis(10,1) = inv_BK(1,1)*grad_basis_temp(10,1) + inv_BK(1,2)*grad_basis_temp(10,2);
		grad_basis(10,2) = inv_BK(2,1)*grad_basis_temp(10,1) + inv_BK(2,2)*grad_basis_temp(10,2);

	}

	free(inv_BK);
	free(grad_basis_temp);

}



/* invert a 2x2 matrix manually */
#define B(i,j) 	   B[ (j-1) + 2*(i-1) ]
#define inv_B(i,j) inv_B[ (j-1) + 2*(i-1) ]

void get_2x2_inv(double* inv_B,
				 double* B)
{

	double det;

	det = B(1,1)*B(2,2) - B(1,2)*B(2,1);

	inv_B(1,1) = 1/det*B(2,2);
	inv_B(2,2) = 1/det*B(1,1);
	inv_B(1,2) = 1/det*-B(1,2);
	inv_B(2,1) = 1/det*-B(2,1);

}